header {* Very Simple Higher-Order Logic *}
theory VSHOL
imports Pure
begin

typedecl bool

judgment
  Trueprop   :: "bool => prop"                   ("(_)" 5)

axiomatization
  The        :: "('a => bool) => 'a"    (binder "THE " 10)
and
  Equality   :: "['a, 'a] => bool"      (infixl "=" 50)
and
  Imp        :: "[bool, bool] => bool"  (infixr "-->" 25)

axiomatization where
  subst:  "[| s= t; P s |] ==> P t"
axiomatization where
  refl:   "t = t"
axiomatization where
  ext:     "[| !! x. (f x = g x) |] ==> (%x. f x) = (%x. g x)"
axiomatization where
  the_eq: "(THE x. x = a) = a"
axiomatization where
  impI:   "(P ==> Q) ==> P-->Q"
axiomatization where
  mp:     "[| P-->Q;  P |] ==> Q"
axiomatization where
  iff:    "(P-->Q) --> (Q-->P) --> (P=Q)"


definition
  True :: bool
where
  "True == ((%x::bool. x) = (%x. x))"

definition
  All :: "('a => bool) => bool"      (binder "ALL " 10)
where
  "All P == (P = (%x. True))"

definition
  Ex :: "('a => bool) => bool"       (binder "EX " 10)
where
  "Ex P == ALL Q. (ALL x. P x --> Q) --> Q"

definition
  False :: bool
where
"False == (ALL P. P)"

definition
  Not :: "bool => bool"              ("~ _" [40] 40)
where
  "~ P == P--> False"

definition
  And :: "[bool, bool] => bool"   (infixr "&" 35)
where
  "P & Q == ALL R. (P-->Q-->R) --> R"

definition
  Or :: "[bool, bool] => bool"   (infixr "|" 30)
where
  "P | Q == ALL R. (P-->R) --> (Q-->R) --> R"

axiomatization where
  True_or_False:  "(P=True) | (P=False)"

text{*
  Abgeleitete Theoreme. 
  *}

theorem sym: "[| a= b |] ==> b= a"
  apply (rule subst [where P="%x. x= a"]) 
  apply (assumption)
  apply (rule refl)
  done

theorem trans: "[| s= t; t= u |] ==> s= u"
  apply (rule subst [where P="%x. x= u" and s="t"])
  apply (rule sym)
  apply (assumption)
  apply (assumption)
  done

lemma ssubst: "[| s= t; P t |] ==> P s"
  apply (rule subst [where P="P"])
  apply (rule sym)
  apply (assumption)
  apply (assumption)
  done

theorem cong: "[| f= g; x= y |] ==> f x = g y"
  apply (rule subst [where P="%a. f x = g a" and s="x"])
  apply (assumption)  
  apply (rule subst [where P="%a. f x = a x" and s="f"])
  apply (assumption)
  apply (rule refl)
  done

theorem arg_cong: "x = y ==> f x = f y"
  apply (rule cong [where f= "f"])
  apply (rule refl)
  apply (assumption)
  done

theorem iffI: "[| P==> Q; Q ==> P |] ==> P= Q"
  apply (rule mp [where P="Q--> P"])
  apply (rule mp [where P="P--> Q"])
  apply (rule iff)
  apply (rule impI) 
  apply (assumption)
  apply (rule impI)
  apply (assumption)
  done

theorem iffD2: "[| P= Q; Q |] ==> P"
  apply (rule subst [where s= "Q"])
  apply (rule sym)
  apply (assumption)
  apply (assumption)
  done  

theorem true: "True"
  apply (unfold True_def)
  apply (rule refl)
  done

theorem eqTrueI: "P ==> P= True"
  apply (rule iffI)
  apply (rule true)
  apply (assumption)
  done

theorem eqTrueE: "P= True ==> P"
  apply (rule iffD2 [where Q="True"])
  apply (assumption)
  apply (rule true)
  done

theorem allI: "[| !!x. P x |] ==> ALL x. P x"
  apply (unfold All_def)
  apply (rule ext)
  apply (rule eqTrueI)
  apply (assumption)
  done

theorem spec: "[| ALL x. P x |] ==> P t"
  apply (unfold All_def)
  apply (rule eqTrueE)
  apply (rule cong [where f="P" and g="%x. True" and x="t"])
  apply (assumption)
  apply (rule refl)
  done

lemma disjE: "[| P | Q; P==> R; Q==> R |] ==> R"
  apply (unfold Or_def)
  apply (rule mp [where P="Q--> R"])
  apply (rule mp [where P="P--> R"])
  apply (rule spec [where P="%R. (P --> R) --> (Q --> R) --> R"])
  apply (assumption)
  apply (rule impI)
  apply (assumption)
  apply (rule impI)
  apply (assumption)
  done

lemma conjI: "[| P ; Q |] ==> P & Q"
  apply (unfold And_def)
  apply (rule allI)
  apply (rule impI)
  apply (rule mp [where P="Q"])
  apply (rule mp [where P="P"])
  apply (assumption)+
  done


text{*
  Function composition.
  *}

definition
  compose :: "('b=> 'c)=> ('a => 'b)=> 'a=> 'c"  (infix "o" 55)
 where
  "(f o g) x == f (g x)"
  
lemma comp_assoc: "(f o g) o h = f o (g o h)"
  apply (rule ext)
  apply (unfold compose_def)
  apply (rule refl)
  done

definition 
  id :: "'a=> 'a"
where
  "id x == x"

lemma "id o f = f"
  apply (rule ext)
  apply (unfold compose_def)
  apply (unfold id_def)
  apply (rule refl)
  done

lemma id_r: "f o id = f"
  apply (rule ext)
  apply (unfold compose_def id_def)
  apply (rule refl)
  done

end

